"""
This is the module that contains the main transformation from ufl to loopy
(with pdelab as the hardcoded generation target)
"""

from __future__ import absolute_import

from dune.perftool import Restriction
from dune.perftool.ufl.modified_terminals import ModifiedTerminalTracker
from dune.perftool.pymbolic.uflmapper import UFL2PymbolicMapper
from dune.perftool.pdelab.geometry import GeometryMapper
from dune.perftool.generation import (domain,
                                      global_context,
                                      globalarg,
                                      iname,
                                      instruction,
                                      symbol,
                                      temporary_variable,
                                      valuearg,
                                      )

from dune.perftool.pdelab.basis import (lfs_iname,
                                        name_lfs,
                                        name_lfs_bound,
                                        traverse_lfs_tree,
                                        )


@iname
def index_sum_iname(i):
    from dune.perftool.pdelab import name_index
    return name_index(i)


class UFL2LoopyVisitor(ModifiedTerminalTracker, UFL2PymbolicMapper, GeometryMapper):
    def __init__(self):
        super(UFL2LoopyVisitor, self).__init__()

    def coefficient(self, o):
        # All trial functions should already be handled
        assert o.count() != 0

        # We expect all coefficients to be of type Expression!
        from dune.perftool.ufl.execution import Expression
        assert isinstance(o, Expression)

        # Determine the name of the parameter function
        from dune.perftool.generation import get_global_context_value
        name = get_global_context_value("namedata")[id(o)]

        # Trigger the generation of code for this thing in the parameter class
        from dune.perftool.pdelab.parameter import (cell_parameter_function,
                                                    intersection_parameter_function,
                                                    )
        if o.on_intersection:
            intersection_parameter_function(name, o)
        else:
            cell_parameter_function(name, o, self.restriction)

        # And return a symbol
        from pymbolic.primitives import Variable
        return Variable(name)

    def index_sum(self, o):
        from dune.perftool.ufl.shape import determine_shape

        # Define an iname for each of the indices in the multiindex
        for i in o.ufl_operands[1].indices():
            shape = determine_shape(o.ufl_operands[0], i)
            index_sum_iname(i)
            from dune.perftool.pdelab import name_index
            domain(name_index(i), shape)

        # Now continue processing the expression
        return self.call(o.ufl_operands[0])


class _Counter:
    counter = 0


def get_count():
    c = _Counter.counter
    _Counter.counter = c + 1
    return c


def transform_accumulation_term(term, measure, subdomain_id):
    from dune.perftool.ufl.transformations.replace import ReplaceExpression
    from pymbolic.primitives import Variable

    # We always have a quadrature loop
    from dune.perftool.pdelab.quadrature import quadrature_iname
    quadrature_iname()

    # Get the pymbolic expression needed for this accumulation term.
    # This includes filling the cache with all sorts of necessary preambles!
    from dune.perftool.ufl.modified_terminals import extract_modified_arguments
    test_ma = extract_modified_arguments(term, trialfunction=False, testfunction=True)
    trial_ma = extract_modified_arguments(term, trialfunction=True, testfunction=False)

    rmap = {}
    for ma in test_ma:
        # If this is a boundary integral, all terms are implicitly restricted to the inside cell
        if measure == 'exterior_facet':
            ma.restriction = Restriction.NEGATIVE

        # Set up the local function space structure
        traverse_lfs_tree(ma)

        # Get the expression for the modified argument representing the test function
        from dune.perftool.pdelab.argument import pymbolic_testfunction
        rmap[ma.expr] = pymbolic_testfunction(ma)

    for ma in trial_ma:
        # If this is a boundary integral, all terms are implicitly restricted to the inside cell
        if measure == 'exterior_facet':
            ma.restriction = Restriction.NEGATIVE

        # Set up the local function space structure
        traverse_lfs_tree(ma)

        # Get the expression for the modified argument representing the trial function
        from dune.perftool.pdelab.argument import pymbolic_trialfunction
        rmap[ma.expr] = pymbolic_trialfunction(ma)

    # Get the transformer!
    ufl2l_mf = UFL2LoopyVisitor()
    re_mf = ReplaceExpression(replacemap=rmap, otherwise=ufl2l_mf)
    ufl2l_mf.call = re_mf.__call__

    pymbolic_expr = re_mf(term)

    # Now simplify the expression
    # TODO: Add a switch to disable/configure this.
    from dune.perftool.pymbolic.simplify import simplify_pymbolic_expression
    pymbolic_expr = simplify_pymbolic_expression(pymbolic_expr)

    # There are some corner cases in DG, where this accumulation term vanishes completely
    if pymbolic_expr == 0:
        return

    # Define a temporary variable for this expression
    expr_tv_name = "expr_" + str(get_count()).zfill(4)
    from pymbolic.primitives import Variable

    # This is a bit hacky now. To correctly determine the iname dependencies of
    # the accumulation term, we inspect it manually. This is necessary, as loopys
    # automatic detection would also considers those inames dependencies which are
    # duplicates of the lfs inames.
    from dune.perftool.pymbolic.inameset import get_index_inames
    acc_inames = get_index_inames(pymbolic_expr).union(frozenset({quadrature_iname()}))

    insn_id = instruction(assignee=Variable(expr_tv_name),
                          expression=pymbolic_expr,
                          forced_iname_deps=acc_inames,
                          forced_iname_deps_is_final=True,
                          )

    # Actually, if we have a cache hit, we need to change our temporary
    from dune.perftool.generation import retrieve_cache_items
    expr_tv_name = filter(lambda i: i.id == insn_id, retrieve_cache_items("instruction"))[0].assignee_name

    # Now register the temporary variable with loopy
    expr_tv = temporary_variable(expr_tv_name)

    # The data that is used to collect the arguments for the accumulate function
    accumargs = []
    residual_shape = {}
    arg_restr = [None] * len(test_ma)

    # Determine whether the inames in the jacobian case clash. This is the case, when
    # we have a galerkin method.
    galerkin = False
    if len(test_ma) == 2:
        galerkin = (test_ma[0].argexpr.element() == test_ma[1].argexpr.element())

    # Generate the code for the modified arguments:
    for arg in test_ma:
        from dune.perftool.pdelab.argument import pymbolic_argument
        from dune.perftool.pdelab.basis import name_lfs
        accumargs.append(name_lfs(arg.argexpr.element()))

        if galerkin and arg.argexpr.count() == 1:
            iname = lfs_iname(arg.argexpr.element(), context="arg")
        else:
            iname = lfs_iname(arg.argexpr.element())

        accumargs.append(iname)

        # Add this iname to the set of used inames for dependencies later on
        acc_inames = acc_inames.union(frozenset({iname}))

        # Determine the shape
        residual_shape[arg.argexpr.number()] = name_lfs_bound(name_lfs(arg.argexpr.element()))

        # Determine the restriction for later
        arg_restr[arg.argexpr.count()] = arg.restriction

    from dune.perftool.pdelab.argument import name_accumulation_variable
    accumvar = name_accumulation_variable(arg_restr)

    # The residual/the jacobian should be represented through a loopy global argument
    # TODO this seems still a bit hacky, esp. w.r.t. systems
    shape = tuple(v for k, v in sorted(residual_shape.items(), key=lambda (k, v): k))
    globalarg(accumvar, shape=shape)

    from dune.perftool.pdelab.quadrature import name_factor
    factor = name_factor()

    # Generate the code snippet for this accumulation instruction
    code = "{}.accumulate({}, {}*{});".format(accumvar,
                                              ", ".join(accumargs),
                                              expr_tv_name,
                                              factor,
                                              )

    # Maybe wrap this instruction into a condiditional. This mostly happens with mixed boundary conditions
    if subdomain_id not in ['everywhere', 'otherwise']:
        # We need to reconstruct the subdomain_data parameter of the measure
        # I am *totally* confused as to why this information is not at hand anyway,
        # but conversation with Martin pointed me to dolfin.fem.assembly where this
        # is done in preprocessing with the limitation of only one possible type of
        # modified measure per integral type.

        # Get the original form and inspect the present measures
        from dune.perftool.generation import get_global_context_value
        original_form = get_global_context_value("formdata").original_form
        sd = original_form.subdomain_data()
        assert len(sd) == 1
        subdomains, = list(sd.values())
        domain, = list(sd.keys())
        for k in list(subdomains.keys()):
            if subdomains[k] is None:
                del subdomains[k]

        # Finally extract the original subdomain_data (which needs to be present!)
        assert measure in subdomains
        subdomain_data = subdomains[measure]

        # Determine the name of the parameter function
        name = get_global_context_value("namedata")[id(subdomain_data)]

        # Trigger the generation of code for this thing in the parameter class
        from dune.perftool.pdelab.parameter import intersection_parameter_function
        intersection_parameter_function(name, subdomain_data, t='int')

        code = "if ({} == {})\n  {}".format(name, subdomain_id, code)

    # Finally, issue the instruction
    instruction(code=code,
                assignees=frozenset({accumvar}),
                read_variables=frozenset({accumvar, factor, expr_tv_name}),
                forced_iname_deps=acc_inames,
                forced_iname_deps_is_final=True,
                )
